
/*  m_avg calculates the average of a set of data points held in a resizeable circular buffer.
It also provides minimum and maximum values in the buffer -- for no really good reason.

Copyright 2007 Andrew May.
Permission is granted to use this software for any purpose provided you
keep this copyright notice intact.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

Updates available from http://cemi.music.unt.edu/may

the author asserts that this may be of interest to computer
musicians wishing to analyze data about musical parameters
such as pitch, amplitude, time, spectral density, and so on,
and furthermore that it provides possibly useful measures
of the tendencies of those data

related externs include:
m_para, m_line, f_avg, f_line, f_para, circ_buff

Note: to substitute for MaxFTS or MaxMSP version, must create abstraction adjust inputs and outputs...
*/

#include "m_pd.h"
#include <stdlib.h>
#include <math.h>

static t_class *m_avg_class; /* instance structure */

typedef struct _node {
	long double	val;
	struct node* next;
	} node;
typedef struct _m_avg
	{
	t_object        x_obj;		          /* obligatory object header */
	
	/* outlets */
	
	t_outlet *out_a;  
	t_outlet *out_err;
	t_outlet *out_min;  
	t_outlet *out_max;  
	t_outlet *out_size;
	
	/* output data */
	
	long double AVG;
	long double ERR;
	
	long double MIN;
	long double MAX;

	float SIZE;	/* number of elements, or n */
	
	/* maximum number of elements in buffer */
	
	float BUFF_SIZE; 

	/* pointers to input data buffer */

	node* head;
	node* tail;
	
	long double val;
	long double E2;
	long double E;

	} t_m_avg;

static void m_avg_free(t_m_avg *x)
{
node* current; 
node* next;

current = x->head; /* copy pointer to head of list into temporary variable */
while (current != NULL) 
	{
	next = current->next;	/* note the next pointer */
	free(current);			/* delete the node */
	current = next;			/* advance to the next node */
	}
x->head = NULL;
}

static void m_avg_clear(t_m_avg *x)
{
if ( x->SIZE > 0) outlet_float(x->out_size, 0); /* if a post-load reset, send reset size = 0 */
x->E = 0;
x->E2 = 0;
x->AVG = 0;
x->ERR = 0;
x->MIN = 0;
x->MAX = 0;
x->SIZE = 0;
m_avg_free(x);
}

static void E_add (t_m_avg *x)
{
x->E2 += x->val * x->val;
x->E += x->val;
}

static void E_subtract (t_m_avg *x)
{
x->E2 -= x->val * x->val;
x->E -= x->val;
}

static void m_avg_minmax (t_m_avg *x)
{
node* current;

if (x->head != NULL)
	{
	current = x->head;
	x->MIN = current->val;
	x->MAX = current->val;
	while (current != NULL)
		{
		if (current->val < x->MIN) x->MIN = current->val;
			else if (current->val > x->MAX) x->MAX = current->val;
		current = current->next;
		}
outlet_float(x->out_max, x->MAX); 
outlet_float(x->out_min, x->MIN);
	}
}

static void m_avg_calc(t_m_avg *x)	/* calculate the moving average */
{

if (x->SIZE >= 1)
	{    
	x->AVG = x->E / x->SIZE;  

	if (x->SIZE == 1) x->ERR = 0;
		else x->ERR = sqrt ((x->E2 - x->E * x->E / x->SIZE) / x->SIZE);

	/* NAN protect */		
	if (!((x->AVG <= 0) || (x->AVG >= 0))) x->AVG = 0;
	if (!((x->ERR <= 0) || (x->ERR >= 0))) x->ERR = 0;

	outlet_float(x->out_a, x->AVG);
	outlet_float(x->out_err, x->ERR);		
	}
}

static void m_avg_float (t_m_avg *x, t_floatarg f)       /* handle a new value */
{										
node* newNode;
node* temp_head;
t_float newsize;

/* determine size of data population and output it */

if (x->SIZE < x->BUFF_SIZE) newsize = x->SIZE + 1;
	else newsize = x->SIZE;
outlet_float(x->out_size, newsize); 

/* allocate memory for new node and link it at the tail of the list */

newNode = malloc(sizeof(node));
x->val = (long double)f;		/* grab the new datum from the input list */
newNode->val = x->val;			/* add them as a node in the linked list */
if (x->SIZE == 0) 
	{				
	newNode->next = NULL;
	x->head = newNode;
	x->tail = x->head;							/* first new pointer is both head and tail */
	x->MIN = x->val;
	x->MAX = x->val;
	}
else
	{
	x->tail->next = newNode;					/* subsequent values push the tail onward */
	x->tail = newNode;
	x->tail->next = NULL;
	}
E_add(x);										/* add the new data to the running sums */

/* make this a circular buffer by getting rid of the the first element if it's full */

if (x->SIZE == x->BUFF_SIZE)
	{
	temp_head = x->head;
	x->val = temp_head->val;		/* grab the new data from the input list */
	x->head = temp_head->next;	/* assign head to the second element */
	free(temp_head);			/* free the memory for the former first element */
	E_subtract(x);
	}
x->SIZE = newsize;
m_avg_minmax(x);	/* run through all remaining x values in list and find new MIN_X and MAX_X */
m_avg_calc(x);
}

static void m_avg_bufsize (t_m_avg *x, t_floatarg n)
{
node* temp_head;
int i;
int deleted;

/* if new buffer size is less than 1, make it 1.
If new buffer size is less than the current data population size, cut the oldest (SIZE - n) from the list
and subtract them from the running sums.
If the new buffer size is greater than the current data population size, change the buffer size
and wait for the buffer to fill with new data. */

if (n<1) 
	{
	post ("m_avg: buffer size set to minimum = 1");
	n=1;
	}
x->BUFF_SIZE = n;
if (n<x->SIZE) {
	deleted = x->SIZE - n;
	for (i = 0; i < deleted; i++)
		{
		temp_head = x->head;
		x->val = x->head->val;	/* data to be returned -- get it before the node is deleted */
		x->head = x->head->next;	/* cut the first element loose by assigning the head pointer to the second element */
		free(temp_head);			/* free the former head node */
		E_subtract(x); /* do the math: subtract the values of the deleted node from running sums */
		x->SIZE -= 1;
		}
m_avg_minmax(x); /* run through all remaining x values in list and find new MIN and MAX */
	}
}

void *m_avg_new(t_floatarg n)	     
{

/* creation argument 1: buffer size (minimum 3) ... */

t_m_avg *x = (t_m_avg *)pd_new(m_avg_class);

x->BUFF_SIZE = 1;
	if (n < 1) post ("m_avg: buffer size set to minimum = 1");
	else x->BUFF_SIZE = n;
		
x->out_a = outlet_new(&x->x_obj, gensym("float"));						/* setup outlets */
x->out_err = outlet_new(&x->x_obj, gensym("float"));
x->out_min = outlet_new(&x->x_obj, gensym("float"));						
x->out_max = outlet_new(&x->x_obj, gensym("float"));
x->out_size = outlet_new(&x->x_obj, gensym("float"));
inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("size"));	/* setup right 2 inlets*/ 
	
x->head = NULL;				/* tidy up initial values */
x->tail = NULL;
m_avg_clear (x); 
	
return (x);
}

void m_avg_setup(void)
{
m_avg_class = class_new(gensym("m_avg"), (t_newmethod)m_avg_new,
	(t_method)m_avg_free, sizeof(t_m_avg), CLASS_DEFAULT, A_DEFFLOAT, 0);

class_addbang(m_avg_class, m_avg_clear);
class_addfloat(m_avg_class, m_avg_float);
class_addmethod(m_avg_class, (t_method)m_avg_bufsize, gensym("size"), A_DEFFLOAT, 0);
}
